import json
import os
import time
from datetime import datetime
from datetime import timezone
from typing import List, Optional, Dict, Union

import pytest
import uuid

import weaviate
from weaviate import Tenant
from weaviate.gql.get import LinkTo


def get_query_for_group(name):
    return (
        """
    {
      Get {
        Group (where: {
          path: ["name"]
          operator: Equal
          valueText: "%s"
        }) {
          name
          _additional {
            id
          }
          members {
            ... on Person {
              name
              _additional {
                id
              }
            }
          }
        }
      }
    }
    """
        % name
    )


gql_get_sophie_scholl = """
{
  Get {
    Person (where: {
      path: ["id"]
      operator: Equal
      valueString: "594b7827-f795-40d0-aabb-5e0553953dad"
    }){
      name
      _additional {
        id
      }
    }
  }
}
"""

SHIP_SCHEMA = {
    "classes": [
        {
            "class": "Ship",
            "description": "object",
            "properties": [
                {"dataType": ["string"], "description": "name", "name": "name"},
                {"dataType": ["string"], "description": "description", "name": "description"},
                {"dataType": ["int"], "description": "size", "name": "size"},
            ],
        }
    ]
}


@pytest.fixture(scope="module")
def people_schema() -> str:
    with open(os.path.join(os.path.dirname(__file__), "people_schema.json"), encoding="utf-8") as f:
        return json.load(f)


def test_load_scheme(people_schema):
    client = weaviate.Client("http://localhost:8080")
    client.schema.delete_all()
    client.schema.create(people_schema)

    assert client.schema.contains()
    assert client.schema.contains(people_schema)

    for cls in people_schema["classes"]:
        client.schema.delete_class(cls["class"])


@pytest.fixture(scope="module")
def client(people_schema):
    client = weaviate.Client("http://localhost:8080")
    client.schema.delete_all()
    client.schema.create(people_schema)

    yield client
    client.schema.delete_all()


@pytest.mark.parametrize("timeout, error", [(None, TypeError), ((5,), ValueError)])
def test_timeout_error(timeout, error):
    with pytest.raises(error):
        weaviate.Client("http://localhost:8080", timeout_config=timeout)


@pytest.mark.parametrize("timeout", [(5, 5), 5, 5.0, (5.0, 5.0), (5, 5.0)])
def test_timeout(people_schema, timeout):
    client = weaviate.Client("http://localhost:8080", timeout_config=timeout)
    client.schema.delete_all()
    client.schema.create(people_schema)
    expected_name = "Sophie Scholl"
    client.data_object.create(
        {"name": expected_name}, "Person", "594b7827-f795-40d0-aabb-5e0553953dad"
    )
    time.sleep(0.5)
    result = client.query.raw(gql_get_sophie_scholl)
    assert result["data"]["Get"]["Person"][0]["name"] == expected_name
    client.schema.delete_all()


@pytest.mark.parametrize("limit", [None, 1, 5, 20, 50])
def test_query_get_with_limit(people_schema, limit: Optional[int]):
    client = weaviate.Client("http://localhost:8080")
    client.schema.delete_all()
    client.schema.create(people_schema)

    num_objects = 20
    for i in range(num_objects):
        with client.batch as batch:
            batch.add_data_object({"name": f"name{i}"}, "Person")
        batch.flush()
    result = client.data_object.get(class_name="Person", limit=limit)
    if limit is None or limit >= num_objects:
        assert len(result["objects"]) == num_objects
    else:
        assert len(result["objects"]) == limit
    client.schema.delete_all()


def test_query_get_with_after(people_schema):
    client = weaviate.Client("http://localhost:8080")
    client.schema.delete_all()
    client.schema.create(people_schema)

    num_objects = 20
    for i in range(num_objects):
        with client.batch as batch:
            batch.add_data_object({"name": f"name{i}"}, "Person")
        batch.flush()

    full_results = client.data_object.get(class_name="Person")
    for i, person in enumerate(full_results["objects"][:-1]):
        results = client.data_object.get(class_name="Person", limit=1, after=person["id"])
        assert full_results["objects"][i + 1]["id"] == results["objects"][0]["id"]

    client.schema.delete_all()


@pytest.mark.parametrize("offset", [None, 0, 1, 5, 20, 50])
def test_query_get_with_offset(people_schema, offset: Optional[int]):
    client = weaviate.Client("http://localhost:8080")
    client.schema.delete_all()
    client.schema.create(people_schema)

    num_objects = 20
    for i in range(num_objects):
        with client.batch as batch:
            batch.add_data_object({"name": f"name{i}"}, "Person")
        batch.flush()
    result_without_offset = client.data_object.get(class_name="Person")
    result_with_offset = client.data_object.get(class_name="Person", offset=offset)

    if offset is None:
        assert result_with_offset["objects"] == result_without_offset["objects"]
    elif offset >= num_objects:
        assert len(result_with_offset["objects"]) == 0
    else:
        assert result_with_offset["objects"][:] == result_without_offset["objects"][offset:]
    client.schema.delete_all()


@pytest.mark.parametrize(
    "sort,expected",
    [
        (
            {"properties": "name", "order_asc": True},
            ["name" + "{:02d}".format(i) for i in range(0, 20)],
        ),
        (
            {"properties": "name", "order_asc": False},
            ["name" + "{:02d}".format(i) for i in range(19, -1, -1)],
        ),
        (
            {"properties": ["name"], "order_asc": [False]},
            ["name" + "{:02d}".format(i) for i in range(19, -1, -1)],
        ),
        (
            {"properties": ["description", "size", "name"], "order_asc": [False, True, False]},
            ["name05", "name00", "name06", "name01"],
        ),
        (
            {"properties": ["description", "size", "name"], "order_asc": False},
            ["name09", "name04", "name08", "name03"],
        ),
        (
            {"properties": ["description", "size", "name"], "order_asc": True},
            ["name10", "name15", "name11", "name16"],
        ),
        ({"properties": ["description", "size", "name"]}, ["name10", "name15", "name11", "name16"]),
    ],
)
def test_query_get_with_sort(
    sort: Optional[Dict[str, Union[str, bool, List[bool], List[str]]]], expected: List[str]
):
    client = weaviate.Client("http://localhost:8080")
    client.schema.delete_all()
    client.schema.create(SHIP_SCHEMA)

    num_objects = 10
    for i in range(num_objects):
        with client.batch as batch:
            batch.add_data_object(
                {
                    "name": "name" + "{:02d}".format(i),
                    "size": i % 5 + 5,
                    "description": "Super long description",
                },
                "Ship",
            )
            batch.add_data_object(
                {
                    "name": "name" + "{:02d}".format(i + 10),
                    "size": i % 5 + 5,
                    "description": "Short description",
                },
                "Ship",
            )
        batch.flush()
    result = client.data_object.get(class_name="Ship", sort=sort)

    for i, exp in enumerate(expected):
        assert exp == result["objects"][i]["properties"]["name"]
    client.schema.delete_all()


def test_query_data(client: weaviate.Client):
    expected_name = "Sophie Scholl"
    client.data_object.create(
        {"name": expected_name}, "Person", "594b7827-f795-40d0-aabb-5e0553953dad"
    )
    time.sleep(2.0)
    result = client.query.raw(gql_get_sophie_scholl)
    assert result["data"]["Get"]["Person"][0]["name"] == expected_name


def test_create_schema():
    client = weaviate.Client("http://localhost:8080")
    single_class = {
        "class": "Barbecue",
        "description": "Barbecue or BBQ where meat and vegetables get grilled",
    }
    client.schema.create_class(single_class)
    prop = {
        "dataType": ["string"],
        "description": "how hot is the BBQ in C",
        "name": "heat",
    }
    client.schema.property.create("Barbecue", prop)
    classes = client.schema.get()["classes"]
    found = False
    for class_ in classes:
        if class_["class"] == "Barbecue":
            found = len(class_["properties"]) == 1
    assert found
    client.schema.delete_class("Barbecue")


def test_replace_and_update(client: weaviate.Client):
    """Test updating an object with put (replace) and patch (update)."""
    uuid = "28954264-0449-57a2-ade5-e9e08d11f51a"
    client.data_object.create({"name": "Someone"}, "Person", uuid)
    person = client.data_object.get_by_id(uuid, class_name="Person")
    assert person["properties"]["name"] == "Someone"
    client.data_object.replace({"name": "SomeoneElse"}, "Person", uuid)
    person = client.data_object.get_by_id(uuid, class_name="Person")
    assert person["properties"]["name"] == "SomeoneElse"
    client.data_object.update({"name": "Anyone"}, "Person", uuid)
    person = client.data_object.get_by_id(uuid, class_name="Person")
    assert person["properties"]["name"] == "Anyone"
    client.data_object.delete(uuid, class_name="Person")


def test_crud(client: weaviate.Client):
    chemists: List[str] = []
    _create_objects_batch(client)
    _create_objects(client, chemists)
    time.sleep(2.0)
    _create_references(client, chemists)
    time.sleep(2.0)
    _validate_data_loading(client)
    _delete_objects(client, chemists)

    _delete_references(client)
    _get_data(client)


def _create_objects_batch(local_client: weaviate.Client):
    local_client.batch.add_data_object({"name": "John Rawls"}, "Person")
    local_client.batch.add_data_object({"name": "Emanuel Kant"}, "Person")
    local_client.batch.add_data_object({"name": "Plato"}, "Person")
    local_client.batch.add_data_object({"name": "Pull-up"}, "Exercise")
    local_client.batch.add_data_object({"name": "Squat"}, "Exercise")
    local_client.batch.add_data_object({"name": "Star jump"}, "Exercise")

    local_client.batch.create_objects()


def _create_objects(local_client: weaviate.Client, chemists: List[str]):
    local_client.data_object.create(
        {"name": "Andrew S. Tanenbaum"}, "Person", "28954261-0449-57a2-ade5-e9e08d11f51a"
    )
    local_client.data_object.create(
        {"name": "Alan Turing"}, "Person", "1c9cd584-88fe-5010-83d0-017cb3fcb446"
    )
    local_client.data_object.create(
        {"name": "John von Neumann"}, "Person", "b36268d4-a6b5-5274-985f-45f13ce0c642"
    )
    local_client.data_object.create(
        {"name": "Tim Berners-Lee"}, "Person", "d1e90d26-d82e-5ef8-84f6-ca67119c7998"
    )
    local_client.data_object.create(
        {"name": "Legends"}, "Group", "2db436b5-0557-5016-9c5f-531412adf9c6"
    )
    local_client.data_object.create(
        {"name": "Chemists"}, "Group", "577887c1-4c6b-5594-aa62-f0c17883d9bf"
    )

    for name in ["Marie Curie", "Fritz Haber", "Walter White"]:
        chemists.append(local_client.data_object.create({"name": name}, "Person"))

    local_time = datetime.now(timezone.utc).astimezone()
    local_client.data_object.create(
        {"start": local_time.isoformat()}, "Call", "3ab05e06-2bb2-41d1-b5c5-e044f3aa9623"
    )


def _create_references(local_client: weaviate.Client, chemists: List[str]):
    local_client.data_object.reference.add(
        "2db436b5-0557-5016-9c5f-531412adf9c6",
        "members",
        "b36268d4-a6b5-5274-985f-45f13ce0c642",
        from_class_name="Group",
        to_class_name="Person",
    )
    local_client.data_object.reference.add(
        "2db436b5-0557-5016-9c5f-531412adf9c6",
        "members",
        "1c9cd584-88fe-5010-83d0-017cb3fcb446",
        from_class_name="Group",
        to_class_name="Person",
    )

    for chemist in chemists:
        local_client.batch.add_reference(
            "577887c1-4c6b-5594-aa62-f0c17883d9bf",
            "Group",
            "members",
            chemist,
            to_object_class_name="Person",
        )

    local_client.batch.create_references()


def _validate_data_loading(local_client: weaviate.Client):
    legends = local_client.query.raw(get_query_for_group("Legends"))["data"]["Get"]
    for member in legends["Group"][0]["members"]:
        assert member["name"] in ["John von Neumann", "Alan Turing"]

    group_chemists = local_client.query.raw(get_query_for_group("Chemists"))["data"]["Get"]
    for member in group_chemists["Group"][0]["members"]:
        assert member["name"] in ["Marie Curie", "Fritz Haber", "Walter White"]
    assert len(group_chemists["Group"][0]["members"]) == 3


def _delete_objects(local_client: weaviate.Client, chemists: List[str]):
    local_client.data_object.delete(
        chemists[2], class_name="Person"
    )  # Delete Walter White not a real chemist just a legend
    time.sleep(1.1)
    assert not local_client.data_object.exists(
        chemists[2], class_name="Person"
    ), "Thing was not correctly deleted"


def _delete_references(local_client: weaviate.Client):
    # Test delete reference
    prime_ministers_group = local_client.data_object.create({"name": "Prime Ministers"}, "Group")
    prime_ministers = []
    for name in ["Wim Kok", "Dries van Agt", "Piet de Jong"]:
        prime_ministers.append(local_client.data_object.create({"name": name}, "Person"))
    for prime_minister in prime_ministers:
        local_client.data_object.reference.add(
            prime_ministers_group,
            "members",
            prime_minister,
            from_class_name="Group",
            to_class_name="Person",
        )
    time.sleep(2.0)
    local_client.data_object.reference.delete(
        prime_ministers_group,
        "members",
        prime_ministers[0],
        from_class_name="Group",
        to_class_name="Person",
    )
    time.sleep(2.0)
    prime_ministers_group_object = local_client.data_object.get_by_id(
        prime_ministers_group, class_name="Group"
    )
    assert (
        len(prime_ministers_group_object["properties"]["members"]) == 2
    ), "Reference not deleted correctly"


def _get_data(local_client: weaviate.Client):
    local_client.data_object.create(
        {"name": "George Floyd"}, "Person", "452e3031-bdaa-4468-9980-aed60d0258bf"
    )
    time.sleep(2.0)
    person = local_client.data_object.get_by_id(
        "452e3031-bdaa-4468-9980-aed60d0258bf",
        ["interpretation"],
        with_vector=True,
        class_name="Person",
    )
    assert "vector" in person
    assert (
        "interpretation" in person["additional"]
    ), "additional property _interpretation not in person"

    persons = local_client.data_object.get(with_vector=True)
    assert "vector" in persons["objects"][0], "additional property _vector not in persons"


def test_add_vector_and_vectorizer(client: weaviate.Client):
    """Add objects with and without vector.

    The Vectorizer should create a vector for the object without vector and the given one should be used for the object
    with vector.
    """
    uuid_without_vector = uuid.uuid4()
    uuid_with_vector = uuid.uuid4()
    with client.batch(batch_size=2) as batch:
        batch.add_data_object({"name": "Some Name"}, "Person", uuid=uuid_without_vector)
        batch.add_data_object(
            {"name": "Other Name"}, "Person", uuid=uuid_with_vector, vector=[1] * 300
        )
        batch.flush()

    object_with_vector = client.data_object.get_by_id(
        uuid_with_vector,
        with_vector=True,
        class_name="Person",
    )
    assert object_with_vector["vector"] == [1] * 300

    object_without_vector = client.data_object.get_by_id(
        uuid_without_vector,
        with_vector=True,
        class_name="Person",
    )
    assert object_without_vector["vector"] != [1] * 300


def test_beacon_refs(people_schema: dict):
    client = weaviate.Client("http://localhost:8080")
    client.schema.delete_all()
    client.schema.create(people_schema)

    persons = []
    for i in range(10):
        persons.append(uuid.uuid4())
        client.data_object.create({"name": "randomName" + str(i)}, "Person", persons[-1])

    client.data_object.create({}, "Call", "3ab05e06-2bb2-41d1-b5c5-e044f3aa9623")

    # create refs
    for i in range(5):
        client.data_object.reference.add(
            to_uuid=persons[i],
            from_property_name="caller",
            from_uuid="3ab05e06-2bb2-41d1-b5c5-e044f3aa9623",
            from_class_name="Call",
            to_class_name="Person",
        )

    result = client.query.get(
        "Call",
        [
            "start",
            LinkTo(link_on="caller", linked_class="Person", properties=["name"]),
        ],
    ).do()
    callers = result["data"]["Get"]["Call"][0]["caller"]
    assert len(callers) == 5
    all_names = [caller["name"] for caller in callers]
    assert all("randomName" + str(i) in all_names for i in range(5))


def test_beacon_refs_multiple(people_schema: dict):
    client = weaviate.Client("http://localhost:8080")
    client.schema.delete_all()
    client.schema.create_class(
        {
            "class": "Person",
            "description": "A person such as humans or personality known through culture",
            "properties": [
                {"name": "name", "dataType": ["string"]},
                {"name": "age", "dataType": ["int"]},
                {"name": "born_in", "dataType": ["text"]},
            ],
            "vectorizer": "none",
        }
    )

    client.schema.create_class(
        {
            "class": "Call",
            "description": "A call between two Persons",
            "properties": [
                {"name": "caller", "dataType": ["Person"]},
                {"name": "recipient", "dataType": ["Person"]},
            ],
            "vectorizer": "none",
        }
    )

    persons = []
    for i in range(10):
        persons.append(uuid.uuid4())
        client.data_object.create(
            {"name": "randomName" + str(i), "age": i, "born_in": "city" + str(i)},
            "Person",
            persons[-1],
        )

    call_uuids = [uuid.uuid4(), uuid.uuid4()]
    client.data_object.create({}, "Call", call_uuids[0])
    client.data_object.create({}, "Call", call_uuids[1])

    # create refs
    for i in range(4):
        client.data_object.reference.add(call_uuids[i % 2], "caller", persons[i], "Call", "Person")
        client.data_object.reference.add(
            call_uuids[i % 2], "recipient", persons[i + 5], "Call", "Person"
        )

    result = client.query.get(
        "Call",
        [
            LinkTo(link_on="caller", linked_class="Person", properties=["name", "age"]),
            LinkTo(link_on="recipient", linked_class="Person", properties=["born_in", "age"]),
        ],
    ).do()
    call1 = result["data"]["Get"]["Call"][0]
    call2 = result["data"]["Get"]["Call"][1]

    # each call has two callers and recipients and caller and recipient should contain different entries
    for call in [call1, call2]:
        assert len(call["caller"]) == 2
        assert len(call["recipient"]) == 2

        assert "age" in call["caller"][0] and "name" in call["caller"][0]
        assert "age" in call["recipient"][0] and "born_in" in call["recipient"][0]


def test_beacon_refs_nested():
    client = weaviate.Client("http://localhost:8080")
    client.schema.delete_all()
    client.schema.create_class(
        {
            "class": "A",
            "properties": [{"name": "nonRef", "dataType": ["string"]}],
            "vectorizer": "none",
        }
    )
    client.schema.create_class(
        {
            "class": "B",
            "properties": [
                {"name": "nonRef", "dataType": ["string"]},
                {"name": "refA", "dataType": ["A"]},
            ],
            "vectorizer": "none",
        }
    )
    client.schema.create_class(
        {
            "class": "C",
            "properties": [
                {"name": "nonRef", "dataType": ["string"]},
                {"name": "refB", "dataType": ["B"]},
            ],
            "vectorizer": "none",
        }
    )
    client.schema.create_class(
        {
            "class": "D",
            "properties": [
                {"name": "nonRef", "dataType": ["string"]},
                {"name": "refC", "dataType": ["C"]},
                {"name": "refB", "dataType": ["B"]},
            ],
            "vectorizer": "none",
        }
    )

    uuid_a = client.data_object.create({"nonRef": "A"}, "A")
    uuid_b = client.data_object.create({"nonRef": "B"}, "B")
    client.data_object.reference.add(uuid_b, "refA", uuid_a, "B", "A")

    uuid_c = client.data_object.create({"nonRef": "C"}, "C")
    client.data_object.reference.add(uuid_c, "refB", uuid_b, "C", "B")

    uuid_d = client.data_object.create({"nonRef": "D"}, "D")
    client.data_object.reference.add(uuid_d, "refC", uuid_c, "D", "C")
    client.data_object.reference.add(uuid_d, "refB", uuid_b, "D", "B")

    result = client.query.get(
        "D",
        [
            "nonRef",
            LinkTo(
                link_on="refC",
                linked_class="C",
                properties=[
                    "nonRef",
                    LinkTo(
                        link_on="refB",
                        linked_class="B",
                        properties=[
                            "nonRef",
                            LinkTo(link_on="refA", linked_class="A", properties=["nonRef"]),
                        ],
                    ),
                ],
            ),
            LinkTo(
                link_on="refB",
                linked_class="B",
                properties=[
                    "nonRef",
                    LinkTo(link_on="refA", linked_class="A", properties=["nonRef"]),
                ],
            ),
        ],
    ).do()

    assert result["data"]["Get"]["D"][0]["refC"][0]["refB"][0]["refA"][0]["nonRef"] == "A"
    assert result["data"]["Get"]["D"][0]["refB"][0]["refA"][0]["nonRef"] == "A"


def test_tenants():
    client = weaviate.Client("http://localhost:8080")
    client.schema.delete_all()
    tenants = [
        Tenant(name="tenantA"),
        Tenant(name="tenantB"),
        Tenant(name="tenantC"),
    ]

    class_name_document = "Document"
    client.schema.create_class(
        {
            "class": class_name_document,
            "properties": [
                {"name": "tenant", "dataType": ["text"]},
                {"name": "title", "dataType": ["text"]},
            ],
            "vectorizer": "none",
            "multiTenancyConfig": {"enabled": True},
        }
    )
    client.schema.add_class_tenants(
        class_name=class_name_document,
        tenants=tenants,
    )
    document_uuids = [
        "00000000-0000-0000-0000-000000000011",
        "00000000-0000-0000-0000-000000000022",
        "00000000-0000-0000-0000-000000000033",
    ]
    document_titles = ["GAN", "OpenAI", "SpaceX"]
    for i in range(0, len(document_uuids)):
        client.data_object.create(
            class_name=class_name_document,
            uuid=document_uuids[i],
            data_object={
                "tenant": tenants[i].name,
                "title": document_titles[i],
            },
            tenant=tenants[i].name,
        )
    documents = client.data_object.get(class_name=class_name_document, tenant=tenants[0].name)
    assert len(documents["objects"]) == 1

    class_name_passage = "Passage"
    client.schema.create_class(
        {
            "class": class_name_passage,
            "properties": [
                {"name": "tenant", "dataType": ["text"]},
                {"name": "content", "dataType": ["text"]},
                {"name": "ofDocument", "dataType": ["Document"]},
            ],
            "vectorizer": "none",
            "multiTenancyConfig": {"enabled": True},
        }
    )
    client.schema.add_class_tenants(
        class_name=class_name_passage,
        tenants=tenants,
    )

    passage_uuids = [
        "00000000-0000-0000-0000-000000000001",
        "00000000-0000-0000-0000-000000000002",
        "00000000-0000-0000-0000-000000000003",
    ]
    txts = ["Txt1", "Txt2", "Txt3"]

    for i in range(0, len(passage_uuids)):
        client.data_object.create(
            class_name=class_name_passage,
            uuid=passage_uuids[i],
            data_object={
                "content": txts[i],
                "tenant": tenants[i].name,
            },
            tenant=tenants[i].name,
        )

    for i in range(0, len(passage_uuids)):
        passage = client.data_object.get_by_id(
            class_name=class_name_passage,
            uuid=passage_uuids[i],
            tenant=tenants[i].name,
        )
        assert passage["properties"]["tenant"] == tenants[i].name
        assert passage["properties"]["content"] == txts[i]
    passages = client.data_object.get(class_name=class_name_passage, tenant=tenants[0].name)
    assert len(passages["objects"]) == 1

    for i in range(0, len(passage_uuids)):
        client.data_object.replace(
            class_name=class_name_passage,
            uuid=passage_uuids[i],
            data_object={
                "content": txts[len(txts) - i - 1],
                "tenant": tenants[i].name,
            },
            tenant=tenants[i].name,
        )

    for i in range(0, len(passage_uuids)):
        exists = client.data_object.exists(
            class_name=class_name_passage,
            uuid=passage_uuids[i],
            tenant=tenants[i].name,
        )
        assert exists
        passage = client.data_object.get_by_id(
            class_name=class_name_passage,
            uuid=passage_uuids[i],
            tenant=tenants[i].name,
        )
        assert passage["properties"]["tenant"] == tenants[i].name
        assert passage["properties"]["content"] == txts[len(txts) - i - 1]

    for i in range(0, len(passage_uuids)):
        client.data_object.update(
            class_name=class_name_passage,
            uuid=passage_uuids[i],
            data_object={
                "content": tenants[i].name,
            },
            tenant=tenants[i].name,
        )

    for i in range(0, len(passage_uuids)):
        passage = client.data_object.get_by_id(
            class_name=class_name_passage,
            uuid=passage_uuids[i],
            tenant=tenants[i].name,
        )
        assert passage["properties"]["tenant"] == tenants[i].name
        assert passage["properties"]["content"] == tenants[i].name

    # references
    for i in range(0, len(passage_uuids)):
        client.data_object.reference.add(
            passage_uuids[i],
            "ofDocument",
            document_uuids[i],
            from_class_name="Passage",
            to_class_name="Document",
            tenant=tenants[i].name,
        )

    for i in range(0, len(passage_uuids)):
        passage = client.data_object.get_by_id(
            class_name=class_name_passage,
            uuid=passage_uuids[i],
            tenant=tenants[i].name,
        )
        assert len(passage["properties"]["ofDocument"]) == 1

    for i in range(0, len(passage_uuids)):
        client.data_object.reference.update(
            passage_uuids[i],
            "ofDocument",
            document_uuids[i],
            from_class_name="Passage",
            to_class_names="Document",
            tenant=tenants[i].name,
        )

    for i in range(0, len(passage_uuids)):
        client.data_object.reference.delete(
            passage_uuids[i],
            "ofDocument",
            document_uuids[i],
            from_class_name="Passage",
            to_class_name="Document",
            tenant=tenants[i].name,
        )

    for i in range(0, len(passage_uuids)):
        passage = client.data_object.get_by_id(
            class_name=class_name_passage,
            uuid=passage_uuids[i],
            tenant=tenants[i].name,
        )
        assert len(passage["properties"]["ofDocument"]) == 0

    for i in range(0, len(passage_uuids)):
        client.data_object.delete(
            class_name=class_name_passage,
            uuid=passage_uuids[i],
            tenant=tenants[i].name,
        )

    for i in range(0, len(passage_uuids)):
        exists = client.data_object.exists(
            class_name=class_name_passage,
            uuid=passage_uuids[i],
            tenant=tenants[i].name,
        )
        assert not exists

    for i in range(0, len(document_uuids)):
        client.data_object.delete(
            class_name=class_name_document,
            uuid=document_uuids[i],
            tenant=tenants[i].name,
        )

    for i in range(0, len(document_uuids)):
        exists = client.data_object.exists(
            class_name=class_name_document,
            uuid=document_uuids[i],
            tenant=tenants[i].name,
        )
        assert not exists


@pytest.mark.parametrize(
    "prop_defs,props",
    [
        (
            {
                "dataType": ["text"],
                "name": "name",
            },
            {
                "name": "test",
            },
        ),
        (
            {
                "dataType": ["text[]"],
                "name": "names",
            },
            {
                "names": ["test1", "test2"],
            },
        ),
        (
            {
                "dataType": ["int"],
                "name": "age",
            },
            {
                "age": 42,
            },
        ),
        (
            {
                "dataType": ["int[]"],
                "name": "ages",
            },
            {
                "ages": [42, 43],
            },
        ),
        (
            {
                "dataType": ["number"],
                "name": "height",
            },
            {
                "height": 1.80,
            },
        ),
        (
            {
                "dataType": ["number[]"],
                "name": "heights",
            },
            {
                "heights": [1.00, 1.80],
            },
        ),
        (
            {
                "dataType": ["boolean"],
                "name": "isTall",
            },
            {
                "isTall": True,
            },
        ),
        (
            {
                "dataType": ["boolean[]"],
                "name": "areTall",
            },
            {
                "areTall": [False, True],
            },
        ),
        (
            {
                "dataType": ["date"],
                "name": "birthday",
            },
            {
                "birthday": "2021-01-01T00:00:00Z",
            },
        ),
        (
            {
                "dataType": ["date[]"],
                "name": "birthdays",
            },
            {
                "birthdays": ["2021-01-01T00:00:00Z", "2021-01-02T00:00:00Z"],
            },
        ),
    ],
)
def test_nested_object_datatype(prop_defs: dict, props: dict):
    client = weaviate.Client("http://localhost:8080")
    client.schema.delete_all()
    client.schema.create_class(
        {
            "class": "A",
            "properties": [
                {"name": "nested", "dataType": ["object"], "nestedProperties": [prop_defs]},
            ],
            "vectorizer": "none",
        }
    )

    uuid_ = client.data_object.create({"nested": props}, "A")
    obj = client.data_object.get_by_id(uuid_, class_name="A")
    assert obj["properties"]["nested"] == props



from typing import Any, Dict, List, Optional

from requests.exceptions import ConnectionError as RequestsConnectionError

from weaviate.collections.classes.batch import (
    ErrorReference,
    _BatchDeleteResult,
    _BatchReference,
    BatchReferenceReturn,
)
from weaviate.collections.classes.config import ConsistencyLevel
from weaviate.collections.filters import _FilterToREST
from weaviate.collections.classes.filters import _Filters
from weaviate.connect import Connection
from weaviate.exceptions import UnexpectedStatusCodeException
from weaviate.util import _decode_json_response_dict


class _BatchREST:
    def __init__(
        self, connection: Connection, consistency_level: Optional[ConsistencyLevel]
    ) -> None:
        self.__connection = connection
        self.__consistency_level = consistency_level

    def delete(
        self, class_name: str, where: _Filters, verbose: bool, dry_run: bool, tenant: Optional[str]
    ) -> _BatchDeleteResult:
        payload: Dict[str, Any] = {
            "match": {
                "class": class_name,
                "where": _FilterToREST.convert(where),
            }
        }
        if verbose:
            payload["output"] = "verbose"
        if dry_run:
            payload["dryRun"] = True

        params = {}
        if self.__consistency_level is not None:
            params["consistency"] = self.__consistency_level.value
        if tenant is not None:
            params["tenant"] = tenant

        try:
            response = self.__connection.delete(
                path="/batch/objects",
                weaviate_object=payload,
                params=params,
            )
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError("Batch delete was not successful.") from conn_err
        res = _decode_json_response_dict(response, "Delete in batch")
        assert res is not None
        return _BatchDeleteResult(
            failed=res["results"]["failed"],
            matches=res["results"]["matches"],
            objects=res["results"]["objects"],
            successful=res["results"]["successful"],
        )

    def references(self, references: List[_BatchReference]) -> BatchReferenceReturn:
        params: Dict[str, str] = {}
        if self.__consistency_level is not None:
            params["consistency_level"] = self.__consistency_level

        refs = [
            {"from": ref.from_, "to": ref.to}
            if ref.tenant is None
            else {"from": ref.from_, "to": ref.to, "tenant": ref.tenant}
            for ref in references
        ]

        response = self.__connection.post(
            path="/batch/references", weaviate_object=refs, params=params
        )
        if response.status_code == 200:
            payload = response.json()
            errors = {
                idx: ErrorReference(
                    message=entry["result"]["errors"]["error"][0]["message"],
                    reference=references[idx],
                )
                for idx, entry in enumerate(payload)
                if entry["result"]["status"] == "FAILED"
            }
            return BatchReferenceReturn(
                elapsed_seconds=response.elapsed.total_seconds(),
                errors=errors,
                has_errors=len(errors) > 0,
            )
        raise UnexpectedStatusCodeException("Send ref batch", response)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: v1/weaviate.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from weaviate.proto.v1 import batch_pb2 as v1_dot_batch__pb2
from weaviate.proto.v1 import search_get_pb2 as v1_dot_search__get__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x11v1/weaviate.proto\x12\x0bweaviate.v1\x1a\x0ev1/batch.proto\x1a\x13v1/search_get.proto2\xa0\x01\n\x08Weaviate\x12@\n\x06Search\x12\x1a.weaviate.v1.SearchRequest\x1a\x18.weaviate.v1.SearchReply"\x00\x12R\n\x0c\x42\x61tchObjects\x12 .weaviate.v1.BatchObjectsRequest\x1a\x1e.weaviate.v1.BatchObjectsReply"\x00\x42j\n#io.weaviate.client.grpc.protocol.v1B\rWeaviateProtoZ4github.com/weaviate/weaviate/grpc/generated;protocolb\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "v1.weaviate_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b"\n#io.weaviate.client.grpc.protocol.v1B\rWeaviateProtoZ4github.com/weaviate/weaviate/grpc/generated;protocol"
    _globals["_WEAVIATE"]._serialized_start = 72
    _globals["_WEAVIATE"]._serialized_end = 232
# @@protoc_insertion_point(module_scope)



"""
Connection class definition.
"""
from __future__ import annotations

import datetime
import os
import socket
import time
from threading import Thread, Event
from typing import Any, Dict, Optional, Tuple, Union, cast
from urllib.parse import urlparse

import requests
from authlib.integrations.requests_client import OAuth2Session  # type: ignore
from requests.adapters import HTTPAdapter
from requests.exceptions import ConnectionError as RequestsConnectionError, ReadTimeout
from requests.exceptions import HTTPError as RequestsHTTPError
from requests.exceptions import JSONDecodeError
from pydantic import BaseModel, field_validator, model_validator

from weaviate import __version__ as client_version
from weaviate.auth import AuthCredentials, AuthClientCredentials, AuthApiKey
from weaviate.config import ConnectionConfig
from weaviate.connect.authentication import _Auth
from weaviate.embedded import EmbeddedDB
from weaviate.exceptions import (
    AuthenticationFailedException,
    WeaviateStartUpError,
    WeaviateQueryException,
)
from weaviate.util import (
    _check_positive_num,
    is_weaviate_domain,
    is_weaviate_too_old,
    is_weaviate_client_too_old,
    PYPI_PACKAGE_URL,
    _decode_json_response_dict,
)
from weaviate.warnings import _Warnings
from weaviate.types import NUMBER


try:
    import grpc  # type: ignore
    from grpc import Channel
    from weaviate.proto.v1 import weaviate_pb2_grpc

    has_grpc = True

except ImportError:
    has_grpc = False


JSONPayload = Union[dict, list]
Session = Union[requests.sessions.Session, OAuth2Session]
TIMEOUT_TYPE_RETURN = Tuple[NUMBER, NUMBER]
PYPI_TIMEOUT = 0.1
MAX_GRPC_MESSAGE_LENGTH = 104858000  # 10mb, needs to be synchronized with GRPC server
GRPC_OPTIONS = [
    ("grpc.max_send_message_length", MAX_GRPC_MESSAGE_LENGTH),
    ("grpc.max_receive_message_length", MAX_GRPC_MESSAGE_LENGTH),
]


class ProtocolParams(BaseModel):
    host: str
    port: int
    secure: bool

    @field_validator("host")
    def _check_host(cls, v: str) -> str:
        if v == "":
            raise ValueError("host must not be empty")
        return v

    @field_validator("port")
    def _check_port(cls, v: int) -> int:
        if v < 0 or v > 65535:
            raise ValueError("port must be between 0 and 65535")
        return v


class ConnectionParams(BaseModel):
    http: ProtocolParams
    grpc: Optional[ProtocolParams] = None

    @classmethod
    def from_url(
        cls, url: str, grpc_port: Optional[int] = None, grpc_secure: bool = False
    ) -> ConnectionParams:
        parsed_url = urlparse(url)
        if parsed_url.scheme not in ["http", "https"]:
            raise ValueError(f"Unsupported scheme: {parsed_url.scheme}")
        if parsed_url.port is None:
            port = 443 if parsed_url.scheme == "https" else 80
        else:
            port = parsed_url.port

        return cls(
            http=ProtocolParams(
                host=cast(str, parsed_url.hostname),
                port=port,
                secure=parsed_url.scheme == "https",
            ),
            grpc=ProtocolParams(
                host=cast(str, parsed_url.hostname),
                port=grpc_port,
                secure=grpc_secure or parsed_url.scheme == "https",
            )
            if grpc_port is not None
            else None,
        )

    @classmethod
    def from_params(
        cls,
        http_host: str,
        http_port: int,
        http_secure: bool,
        grpc_host: Optional[str] = None,
        grpc_port: Optional[int] = None,
        grpc_secure: Optional[bool] = None,
    ) -> ConnectionParams:
        return cls(
            http=ProtocolParams(
                host=http_host,
                port=http_port,
                secure=http_secure,
            ),
            grpc=ProtocolParams(
                host=grpc_host,
                port=grpc_port,
                secure=grpc_secure,
            )
            if (grpc_host is not None and grpc_port is not None and grpc_secure is not None)
            else None,
        )

    @model_validator(mode="after")
    def _check_port_collision(self) -> ConnectionParams:
        if self.grpc is not None and self.http.port == self.grpc.port:
            raise ValueError("http.port and grpc.port must be different")
        return self

    @property
    def _grpc_address(self) -> Optional[Tuple[str, int]]:
        if self.grpc is None:
            return None
        return (self.grpc.host, self.grpc.port)

    @property
    def _grpc_target(self) -> Optional[str]:
        if self.grpc is None:
            return None
        return f"{self.grpc.host}:{self.grpc.port}"

    def _grpc_channel(self) -> Optional[Channel]:
        if self.grpc is None:
            return None
        if self.grpc.secure:
            return grpc.secure_channel(
                target=self._grpc_target,
                credentials=grpc.ssl_channel_credentials(),
                options=GRPC_OPTIONS,
            )
        else:
            return grpc.insecure_channel(
                target=self._grpc_target,
                options=GRPC_OPTIONS,
            )

    @property
    def _http_scheme(self) -> str:
        return "https" if self.http.secure else "http"

    @property
    def _http_url(self) -> str:
        return f"{self._http_scheme}://{self.http.host}:{self.http.port}"

    @property
    def _has_grpc(self) -> bool:
        return self.grpc is not None


class Connection:
    """
    Connection class used to communicate to a weaviate instance.
    """

    def __init__(
        self,
        connection_params: ConnectionParams,
        auth_client_secret: Optional[AuthCredentials],
        timeout_config: TIMEOUT_TYPE_RETURN,
        proxies: Union[dict, str, None],
        trust_env: bool,
        additional_headers: Optional[Dict[str, Any]],
        startup_period: Optional[int],
        connection_config: ConnectionConfig,
        embedded_db: Optional[EmbeddedDB] = None,
    ):
        """
        Initialize a Connection class instance.

        Parameters
        ----------
        url : str
            URL to a running weaviate instance.
        auth_client_secret : weaviate.auth.AuthCredentials, optional
            Credentials to authenticate with a weaviate instance. The credentials are not saved within the client and
            authentication is done via authentication tokens.
        timeout_config : tuple(float, float) or float, optional
            Set the timeout configuration for all requests to the Weaviate server. It can be a
            float or, a tuple of two floats: (connect timeout, read timeout).
            If only one float is passed then both connect and read timeout will be set to
            that value.
        proxies : dict, str or None, optional
            Proxies to be used for requests. Are used by both 'requests' and 'aiohttp'. Can be
            passed as a dict ('requests' format:
            https://docs.python-requests.org/en/stable/user/advanced/#proxies), str (HTTP/HTTPS
            protocols are going to use this proxy) or None.
        trust_env : bool, optional
            Whether to read proxies from the ENV variables: (HTTP_PROXY or http_proxy, HTTPS_PROXY
            or https_proxy).
            NOTE: 'proxies' has priority over 'trust_env', i.e. if 'proxies' is NOT None,
            'trust_env' is ignored.
        additional_headers : Dict[str, Any] or None
            Additional headers to include in the requests, used to set OpenAI key. OpenAI key looks
            like this: {'X-OpenAI-Api-Key': 'KEY'}.
        startup_period : int or None
            How long the client will wait for weaviate to start before raising a RequestsConnectionError.
            If None the client will not wait at all.

        Raises
        ------
        ValueError
            If no authentication credentials provided but the Weaviate server has an OpenID
            configured.
        """

        self._api_version_path = "/v1"
        self.url = connection_params._http_url  # e.g. http://localhost:80
        self.timeout_config: TIMEOUT_TYPE_RETURN = timeout_config
        self.embedded_db = embedded_db

        self._grpc_stub: Optional[weaviate_pb2_grpc.WeaviateStub] = None
        self.__additional_headers: Dict[str, str] = {}

        # create GRPC channel. If weaviate does not support GRPC, fallback to GraphQL is used.
        if has_grpc and connection_params._has_grpc:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            try:
                s.settimeout(1.0)  # we're only pinging the port, 1s is plenty
                assert connection_params._grpc_address is not None
                s.connect(connection_params._grpc_address)
                s.shutdown(2)
                s.close()
                grpc_channel = connection_params._grpc_channel()
                assert grpc_channel is not None
                self._grpc_stub = weaviate_pb2_grpc.WeaviateStub(grpc_channel)
                self._grpc_available = True
            except (
                ConnectionRefusedError,
                TimeoutError,
                socket.timeout,
            ):  # self._grpc_stub stays None
                s.close()
                self._grpc_available = False

        self._headers = {"content-type": "application/json"}
        if additional_headers is not None:
            if not isinstance(additional_headers, dict):
                raise TypeError(
                    f"'additional_headers' must be of type dict or None. Given type: {type(additional_headers)}."
                )
            self.__additional_headers = additional_headers
            for key, value in additional_headers.items():
                self._headers[key.lower()] = value

        self._proxies = _get_proxies(proxies, trust_env)

        # auth secrets can contain more information than a header (refresh tokens and lifetime) and therefore take
        # precedent over headers
        if "authorization" in self._headers and auth_client_secret is not None:
            _Warnings.auth_header_and_auth_secret()
            self._headers.pop("authorization")

        # if there are API keys included add them right away to headers
        if auth_client_secret is not None and isinstance(auth_client_secret, AuthApiKey):
            self._headers["authorization"] = "Bearer " + auth_client_secret.api_key

        self._session: Session
        self._shutdown_background_event: Optional[Event] = None

        if startup_period is not None:
            _check_positive_num(startup_period, "startup_period", int, include_zero=False)
            self.wait_for_weaviate(startup_period)

        self._create_session(auth_client_secret)
        self._add_adapter_to_session(connection_config)

        self._server_version = self.get_meta()["version"]
        if self._server_version < "1.14":
            _Warnings.weaviate_server_older_than_1_14(self._server_version)
        if is_weaviate_too_old(self._server_version):
            _Warnings.weaviate_too_old_vs_latest(self._server_version)

        try:
            pkg_info = requests.get(PYPI_PACKAGE_URL, timeout=PYPI_TIMEOUT).json()
            pkg_info = pkg_info.get("info", {})
            latest_version = pkg_info.get("version", "unknown version")
            if is_weaviate_client_too_old(client_version, latest_version):
                _Warnings.weaviate_client_too_old_vs_latest(client_version, latest_version)
        except requests.exceptions.RequestException:
            pass  # ignore any errors related to requests, it is a best-effort warning

    def _create_session(self, auth_client_secret: Optional[AuthCredentials]) -> None:
        """Creates a request session.

        Either through authlib.oauth2 if authentication is enabled or a normal request session otherwise.

        Raises
        ------
        ValueError
            If no authentication credentials provided but the Weaviate server has OpenID configured.
        """
        # API keys are separate from OIDC and do not need any config from weaviate
        if auth_client_secret is not None and isinstance(auth_client_secret, AuthApiKey):
            self._session = requests.Session()
            return

        if "authorization" in self._headers and auth_client_secret is None:
            self._session = requests.Session()
            return

        oidc_url = self.url + self._api_version_path + "/.well-known/openid-configuration"
        response = requests.get(
            oidc_url,
            headers=self._get_request_header(),
            timeout=self._timeout_config,
            proxies=self._proxies,
        )
        if response.status_code == 200:
            # Some setups are behind proxies that return some default page - for example a login - for all requests.
            # If the response is not json, we assume that this is the case and try unauthenticated access. Any auth
            # header provided by the user is unaffected.
            try:
                resp = response.json()
            except JSONDecodeError:
                _Warnings.auth_cannot_parse_oidc_config(oidc_url)
                self._session = requests.Session()
                return

            if auth_client_secret is not None and not isinstance(auth_client_secret, AuthApiKey):
                _auth = _Auth(resp, auth_client_secret, self)
                self._session = _auth.get_auth_session()

                if isinstance(auth_client_secret, AuthClientCredentials):
                    # credentials should only be saved for client credentials, otherwise use refresh token
                    self._create_background_token_refresh(_auth)
                else:
                    self._create_background_token_refresh()
            else:
                msg = f""""No login credentials provided. The weaviate instance at {self.url} requires login credentials.

                    Please check our documentation at https://weaviate.io/developers/weaviate/client-libraries/python#authentication
                    for more information about how to use authentication."""

                if is_weaviate_domain(self.url):
                    msg += """

                    You can instantiate the client with login credentials for WCS using

                    client = weaviate.Client(
                      url=YOUR_WEAVIATE_URL,
                      auth_client_secret=weaviate.AuthClientPassword(
                        username = YOUR_WCS_USER,
                        password = YOUR_WCS_PW,
                      ))
                    """
                raise AuthenticationFailedException(msg)
        elif response.status_code == 404 and auth_client_secret is not None:
            _Warnings.auth_with_anon_weaviate()
            self._session = requests.Session()
        else:
            self._session = requests.Session()

    def get_current_bearer_token(self) -> str:
        if "authorization" in self._headers:
            return self._headers["authorization"]
        elif isinstance(self._session, OAuth2Session):
            return f"Bearer {self._session.token['access_token']}"

        return ""

    def _add_adapter_to_session(self, connection_config: ConnectionConfig) -> None:
        adapter = HTTPAdapter(
            pool_connections=connection_config.session_pool_connections,
            pool_maxsize=connection_config.session_pool_maxsize,
        )
        self._session.mount("http://", adapter)
        self._session.mount("https://", adapter)

    def _create_background_token_refresh(self, _auth: Optional[_Auth] = None) -> None:
        """Create a background thread that periodically refreshes access and refresh tokens.

        While the underlying library refreshes tokens, it does not have an internal cronjob that checks every
        X-seconds if a token has expired. If there is no activity for longer than the refresh tokens lifetime, it will
        expire. Therefore, refresh manually shortly before expiration time is up."""
        assert isinstance(self._session, OAuth2Session)
        if "refresh_token" not in self._session.token and _auth is None:
            return

        expires_in: int = self._session.token.get(
            "expires_in", 60
        )  # use 1minute as token lifetime if not supplied
        self._shutdown_background_event = Event()

        def periodic_refresh_token(refresh_time: int, _auth: Optional[_Auth]) -> None:
            time.sleep(max(refresh_time - 30, 1))
            while (
                self._shutdown_background_event is not None
                and not self._shutdown_background_event.is_set()
            ):
                # use refresh token when available
                try:
                    if "refresh_token" in cast(OAuth2Session, self._session).token:
                        assert isinstance(self._session, OAuth2Session)
                        self._session.token = self._session.refresh_token(
                            self._session.metadata["token_endpoint"]
                        )
                        refresh_time = self._session.token.get("expires_in") - 30
                    else:
                        # client credentials usually does not contain a refresh token => get a new token using the
                        # saved credentials
                        assert _auth is not None
                        new_session = _auth.get_auth_session()
                        self._session.token = new_session.fetch_token()  # type: ignore
                except (RequestsHTTPError, ReadTimeout) as exc:
                    # retry again after one second, might be an unstable connection
                    refresh_time = 1
                    _Warnings.token_refresh_failed(exc)

                time.sleep(max(refresh_time, 1))

        demon = Thread(
            target=periodic_refresh_token,
            args=(expires_in, _auth),
            daemon=True,
            name="TokenRefresh",
        )
        demon.start()

    def close(self) -> None:
        """Shutdown connection class gracefully."""
        # in case an exception happens before definition of these members
        if (
            hasattr(self, "_shutdown_background_event")
            and self._shutdown_background_event is not None
        ):
            self._shutdown_background_event.set()
        if hasattr(self, "_session"):
            self._session.close()

    def _get_request_header(self) -> dict:
        """
        Returns the correct headers for a request.

        Returns
        -------
        dict
            Request header as a dict.
        """
        return self._headers

    def delete(
        self,
        path: str,
        weaviate_object: Optional[JSONPayload] = None,
        params: Optional[Dict[str, Any]] = None,
    ) -> requests.Response:
        """
        Make a DELETE request to the Weaviate server instance.

        Parameters
        ----------
        path : str
            Sub-path to the Weaviate resources. Must be a valid Weaviate sub-path.
            e.g. '/meta' or '/objects', without version.
        weaviate_object : dict, optional
            Object is used as payload for DELETE request. By default None.
        params : dict, optional
            Additional request parameters, by default None

        Returns
        -------
        requests.Response
            The response, if request was successful.

        Raises
        ------
        requests.ConnectionError
            If the DELETE request could not be made.
        """
        if self.embedded_db is not None:
            self.embedded_db.ensure_running()
        request_url = self.url + self._api_version_path + path

        return self._session.delete(
            url=request_url,
            json=weaviate_object,
            headers=self._get_request_header(),
            timeout=self._timeout_config,
            proxies=self._proxies,
            params=params,
        )

    def patch(
        self,
        path: str,
        weaviate_object: JSONPayload,
        params: Optional[Dict[str, Any]] = None,
    ) -> requests.Response:
        """
        Make a PATCH request to the Weaviate server instance.

        Parameters
        ----------
        path : str
            Sub-path to the Weaviate resources. Must be a valid Weaviate sub-path.
            e.g. '/meta' or '/objects', without version.
        weaviate_object : dict
            Object is used as payload for PATCH request.
        params : dict, optional
            Additional request parameters, by default None
        Returns
        -------
        requests.Response
            The response, if request was successful.

        Raises
        ------
        requests.ConnectionError
            If the PATCH request could not be made.
        """
        if self.embedded_db is not None:
            self.embedded_db.ensure_running()
        request_url = self.url + self._api_version_path + path

        return self._session.patch(
            url=request_url,
            json=weaviate_object,
            headers=self._get_request_header(),
            timeout=self._timeout_config,
            proxies=self._proxies,
            params=params,
        )

    def post(
        self,
        path: str,
        weaviate_object: JSONPayload,
        params: Optional[Dict[str, Any]] = None,
    ) -> requests.Response:
        """
        Make a POST request to the Weaviate server instance.

        Parameters
        ----------
        path : str
            Sub-path to the Weaviate resources. Must be a valid Weaviate sub-path.
            e.g. '/meta' or '/objects', without version.
        weaviate_object : dict
            Object is used as payload for POST request.
        params : dict, optional
            Additional request parameters, by default None
        external_url: Is an external (non-weaviate) url called

        Returns
        -------
        requests.Response
            The response, if request was successful.

        Raises
        ------
        requests.ConnectionError
            If the POST request could not be made.
        """
        if self.embedded_db is not None:
            self.embedded_db.ensure_running()
        request_url = self.url + self._api_version_path + path

        return self._session.post(
            url=request_url,
            json=weaviate_object,
            headers=self._get_request_header(),
            timeout=self._timeout_config,
            proxies=self._proxies,
            params=params,
        )

    def put(
        self,
        path: str,
        weaviate_object: JSONPayload,
        params: Optional[Dict[str, Any]] = None,
    ) -> requests.Response:
        """
        Make a PUT request to the Weaviate server instance.

        Parameters
        ----------
        path : str
            Sub-path to the Weaviate resources. Must be a valid Weaviate sub-path.
            e.g. '/meta' or '/objects', without version.
        weaviate_object : dict
            Object is used as payload for PUT request.
        params : dict, optional
            Additional request parameters, by default None
        Returns
        -------
        requests.Response
            The response, if request was successful.

        Raises
        ------
        requests.ConnectionError
            If the PUT request could not be made.
        """
        if self.embedded_db is not None:
            self.embedded_db.ensure_running()
        request_url = self.url + self._api_version_path + path

        return self._session.put(
            url=request_url,
            json=weaviate_object,
            headers=self._get_request_header(),
            timeout=self._timeout_config,
            proxies=self._proxies,
            params=params,
        )

    def get(
        self, path: str, params: Optional[Dict[str, Any]] = None, external_url: bool = False
    ) -> requests.Response:
        """Make a GET request.

        Parameters
        ----------
        path : str
            Sub-path to the Weaviate resources. Must be a valid Weaviate sub-path.
            e.g. '/meta' or '/objects', without version.
        params : dict, optional
            Additional request parameters, by default None
        external_url: Is an external (non-weaviate) url called

        Returns
        -------
        requests.Response
            The response if request was successful.

        Raises
        ------
        requests.ConnectionError
            If the GET request could not be made.
        """
        if self.embedded_db is not None:
            self.embedded_db.ensure_running()
        if params is None:
            params = {}

        if external_url:
            request_url = path
        else:
            request_url = self.url + self._api_version_path + path

        return self._session.get(
            url=request_url,
            headers=self._get_request_header(),
            timeout=self._timeout_config,
            params=params,
            proxies=self._proxies,
        )

    def head(
        self,
        path: str,
        params: Optional[Dict[str, Any]] = None,
    ) -> requests.Response:
        """
        Make a HEAD request to the server.

        Parameters
        ----------
        path : str
            Sub-path to the resources. Must be a valid sub-path.
            e.g. '/meta' or '/objects', without version.
        params : dict, optional
            Additional request parameters, by default None

        Returns
        -------
        requests.Response
            The response to the request.

        Raises
        ------
        requests.ConnectionError
            If the HEAD request could not be made.
        """
        if self.embedded_db is not None:
            self.embedded_db.ensure_running()
        request_url = self.url + self._api_version_path + path

        return self._session.head(
            url=request_url,
            headers=self._get_request_header(),
            timeout=self._timeout_config,
            proxies=self._proxies,
            params=params,
        )

    @property
    def timeout_config(self) -> TIMEOUT_TYPE_RETURN:
        """
        Getter/setter for `timeout_config`.

        Parameters
        ----------
        timeout_config : tuple(float, float), optional
            For Setter only: Set the timeout configuration for all requests to the Weaviate server.
            It can be a float or, a tuple of two floats:
                    (connect timeout, read timeout).
            If only one float is passed then both connect and read timeout will be set to
            that value.

        Returns
        -------
        Tuple[float, float]
            For Getter only: Requests Timeout configuration.
        """

        return self._timeout_config

    @timeout_config.setter
    def timeout_config(self, timeout_config: TIMEOUT_TYPE_RETURN) -> None:
        """
        Setter for `timeout_config`. (docstring should be only in the Getter)
        """

        self._timeout_config = timeout_config

    @property
    def proxies(self) -> dict:
        return self._proxies

    def wait_for_weaviate(self, startup_period: int) -> None:
        """
        Waits until weaviate is ready or the timelimit given in 'startup_period' has passed.

        Parameters
        ----------
        startup_period : int
            Describes how long the client will wait for weaviate to start in seconds.

        Raises
        ------
        WeaviateStartUpError
            If weaviate takes longer than the timelimit to respond.
        """

        ready_url = self.url + self._api_version_path + "/.well-known/ready"
        for _i in range(startup_period):
            try:
                requests.get(ready_url, headers=self._get_request_header()).raise_for_status()
                return
            except (RequestsHTTPError, RequestsConnectionError):
                time.sleep(1)

        try:
            requests.get(ready_url, headers=self._get_request_header()).raise_for_status()
            return
        except (RequestsHTTPError, RequestsConnectionError) as error:
            raise WeaviateStartUpError(
                f"Weaviate did not start up in {startup_period} seconds. Either the Weaviate URL {self.url} is wrong or Weaviate did not start up in the interval given in 'startup_period'."
            ) from error

    @property
    def grpc_stub(self) -> Optional[weaviate_pb2_grpc.WeaviateStub]:
        return self._grpc_stub

    @property
    def server_version(self) -> str:
        """
        Version of the weaviate instance.
        """
        return self._server_version

    @property
    def additional_headers(self) -> Dict[str, str]:
        return self.__additional_headers

    def get_meta(self) -> Dict[str, str]:
        """
        Returns the meta endpoint.
        """
        response = self.get(path="/meta")
        res = _decode_json_response_dict(response, "Meta endpoint")
        assert res is not None
        return res


class GRPCConnection(Connection):
    def __init__(
        self,
        connection_params: ConnectionParams,
        auth_client_secret: Optional[AuthCredentials],
        timeout_config: TIMEOUT_TYPE_RETURN,
        proxies: Union[dict, str, None],
        trust_env: bool,
        additional_headers: Optional[Dict[str, Any]],
        startup_period: Optional[int],
        connection_config: ConnectionConfig,
        embedded_db: Optional[EmbeddedDB] = None,
    ):
        super().__init__(
            connection_params,
            auth_client_secret,
            timeout_config,
            proxies,
            trust_env,
            additional_headers,
            startup_period,
            connection_config,
            embedded_db,
        )
        if self._server_version < "1.21" or self._grpc_stub is None:
            raise WeaviateQueryException(
                f"gRPC is not enabled. Is your Weaviate version at least 1.22 or higher? Current is {self._server_version}"
            )

    @property
    def grpc_stub(self) -> Optional[weaviate_pb2_grpc.WeaviateStub]:
        return self._grpc_stub


def _get_epoch_time() -> int:
    """
    Get the current epoch time as an integer.

    Returns
    -------
    int
        Current epoch time.
    """

    dts = datetime.datetime.utcnow()
    return round(time.mktime(dts.timetuple()) + dts.microsecond / 1e6)


def _get_proxies(proxies: Union[dict, str, None], trust_env: bool) -> dict:
    """
    Get proxies as dict, compatible with 'requests' library.
    NOTE: 'proxies' has priority over 'trust_env', i.e. if 'proxies' is NOT None, 'trust_env'
    is ignored.

    Parameters
    ----------
    proxies : dict, str or None
        The proxies to use for requests. If it is a dict it should follow 'requests' library
        format (https://docs.python-requests.org/en/stable/user/advanced/#proxies). If it is
        a URL (str), a dict will be constructed with both 'http' and 'https' pointing to that
        URL. If None, no proxies will be used.
    trust_env : bool
        If True, the proxies will be read from ENV VARs (case insensitive):
            HTTP_PROXY/HTTPS_PROXY.
        NOTE: It is ignored if 'proxies' is NOT None.

    Returns
    -------
    dict
        A dictionary with proxies, either set from 'proxies' or read from ENV VARs.
    """

    if proxies is not None:
        if isinstance(proxies, str):
            return {
                "http": proxies,
                "https": proxies,
            }
        if isinstance(proxies, dict):
            return proxies
        raise TypeError(
            "If 'proxies' is not None, it must be of type dict or str. "
            f"Given type: {type(proxies)}."
        )

    if not trust_env:
        return {}

    http_proxy = (os.environ.get("HTTP_PROXY"), os.environ.get("http_proxy"))
    https_proxy = (os.environ.get("HTTPS_PROXY"), os.environ.get("https_proxy"))

    if not any(http_proxy + https_proxy):
        return {}

    proxies = {}
    if any(http_proxy):
        proxies["http"] = http_proxy[0] if http_proxy[0] else http_proxy[1]
    if any(https_proxy):
        proxies["https"] = https_proxy[0] if https_proxy[0] else https_proxy[1]

    return proxies



from typing import Dict

from requests.exceptions import ConnectionError as RequestsConnectionError

from weaviate.connect import Connection
from weaviate.collections.classes.config import (
    _CollectionConfigCreateBase,
    _CollectionConfig,
    _CollectionConfigSimple,
)
from weaviate.collections.classes.config_methods import (
    _collection_configs_from_json,
    _collection_configs_simple_from_json,
)
from weaviate.exceptions import UnexpectedStatusCodeException
from weaviate.util import _capitalize_first_letter


class _CollectionBase:
    def __init__(self, name: str) -> None:
        self.name = _capitalize_first_letter(name)


class _CollectionsBase:
    def __init__(self, connection: Connection):
        self._connection = connection

    def _create(
        self,
        config: _CollectionConfigCreateBase,
    ) -> str:
        weaviate_object = config._to_dict()

        try:
            response = self._connection.post(path="/schema", weaviate_object=weaviate_object)
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError("Class may not have been created properly.") from conn_err
        if response.status_code != 200:
            raise UnexpectedStatusCodeException("Create class", response)

        collection_name = response.json()["class"]
        assert isinstance(collection_name, str)
        return collection_name

    def _exists(self, name: str) -> bool:
        path = f"/schema/{name}"
        try:
            response = self._connection.get(path=path)
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError("Existenz of class.") from conn_err

        if response.status_code == 200:
            return True
        elif response.status_code == 404:
            return False
        raise UnexpectedStatusCodeException("collection exists", response)

    def _delete(self, name: str) -> None:
        path = f"/schema/{name}"
        try:
            response = self._connection.delete(path=path)
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError("Deletion of class.") from conn_err
        if response.status_code == 200:
            return

        UnexpectedStatusCodeException("Delete collection", response)

    def _get_all(self) -> Dict[str, _CollectionConfig]:
        try:
            response = self._connection.get(path="/schema")
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError("Get schema.") from conn_err
        if response.status_code == 200:
            res = response.json()
            return _collection_configs_from_json(res)
        raise UnexpectedStatusCodeException("Get schema", response)

    def _get_simple(self) -> Dict[str, _CollectionConfigSimple]:
        try:
            response = self._connection.get(path="/schema")
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError("Get schema.") from conn_err
        if response.status_code == 200:
            res = response.json()
            return _collection_configs_simple_from_json(res)
        raise UnexpectedStatusCodeException("Get schema", response)



from typing import (
    Any,
    Generic,
    List,
    Optional,
    Type,
    Union,
    cast,
    get_type_hints,
)

from weaviate.collections.classes.config import ConsistencyLevel

from weaviate.collections.classes.filters import _Filters
from weaviate.collections.classes.grpc import (
    HybridFusion,
    MetadataQuery,
    Move,
    PROPERTIES,
)
from weaviate.collections.classes.internal import _Object
from weaviate.collections.classes.orm import Model
from weaviate.collections.classes.types import TProperties

from weaviate.collections.data import _DataCollection

from weaviate.collections.grpc.query import SearchResult

from weaviate.collections.queries.base import _Grpc
from weaviate.collections.queries.bm25 import _BM25Generate, _BM25Query
from weaviate.collections.queries.fetch_objects import _FetchObjectsGenerate, _FetchObjectsQuery
from weaviate.collections.queries.hybrid import _HybridGenerate, _HybridQuery
from weaviate.collections.queries.near_audio import (
    _NearAudioGenerate,
    _NearAudioGroupBy,
    _NearAudioQuery,
)
from weaviate.collections.queries.near_image import (
    _NearImageGenerate,
    _NearImageGroupBy,
    _NearImageQuery,
)
from weaviate.collections.queries.near_object import (
    _NearObjectGenerate,
    _NearObjectGroupBy,
    _NearObjectQuery,
)
from weaviate.collections.queries.near_text import (
    _NearTextGenerate,
    _NearTextGroupBy,
    _NearTextQuery,
)
from weaviate.collections.queries.near_vector import (
    _NearVectorGenerate,
    _NearVectorGroupBy,
    _NearVectorQuery,
)
from weaviate.collections.queries.near_video import (
    _NearVideoGenerate,
    _NearVideoGroupBy,
    _NearVideoQuery,
)

from weaviate.connect import Connection
from weaviate.types import UUID

from weaviate.proto.v1 import search_get_pb2


class _QueryCollection(
    Generic[TProperties],
    _BM25Query[TProperties],
    _FetchObjectsQuery[TProperties],
    _HybridQuery,
    _NearAudioQuery,
    _NearImageQuery,
    _NearObjectQuery,
    _NearTextQuery,
    _NearVectorQuery,
    _NearVideoQuery,
):
    def __init__(
        self,
        connection: Connection,
        name: str,
        rest_query: _DataCollection[TProperties],
        consistency_level: Optional[ConsistencyLevel],
        tenant: Optional[str],
        type_: Optional[Type[TProperties]],
    ):
        super().__init__(connection, name, consistency_level, tenant, type_)
        self.__data = rest_query

    def fetch_object_by_id(
        self, uuid: UUID, include_vector: bool = False
    ) -> Optional[_Object[TProperties]]:
        """Retrieve an object from the server by its UUID.

        Arguments:
            `uuid`
                The UUID of the object to retrieve, REQUIRED.
            `include_vector`
                Whether to include the vector in the returned object.

        Raises:
            `weaviate.exceptions.WeaviateQueryException`:
                If the network connection to Weaviate fails.
            `weaviate.exceptions.WeaviateInsertInvalidPropertyError`:
                If a property is invalid. I.e., has name `id` or `vector`, which are reserved.
        """
        ret = self.__data._get_by_id(uuid=uuid, include_vector=include_vector)
        if ret is None:
            return ret
        return self.__data._json_to_object(ret)


class _GenerateCollection(
    _BM25Generate,
    _FetchObjectsGenerate,
    _HybridGenerate,
    _NearAudioGenerate,
    _NearImageGenerate,
    _NearObjectGenerate,
    _NearTextGenerate,
    _NearVectorGenerate,
    _NearVideoGenerate,
):
    pass


class _GroupByCollection(
    _NearAudioGroupBy,
    _NearImageGroupBy,
    _NearObjectGroupBy,
    _NearTextGroupBy,
    _NearVectorGroupBy,
    _NearVideoGroupBy,
):
    pass


class _GrpcCollectionModel(Generic[Model], _Grpc[Any]):
    def __init__(
        self,
        connection: Connection,
        name: str,
        model: Type[Model],
        tenant: Optional[str] = None,
        consistency_level: Optional[ConsistencyLevel] = None,
    ):
        super().__init__(connection, name, consistency_level, tenant, type_=model)
        self.model = model

    def __parse_result(
        self,
        properties: "search_get_pb2.PropertiesResult",
        type_: Type[Model],
    ) -> Model:
        hints = get_type_hints(type_)

        result = {}

        for name, non_ref_prop in properties.non_ref_properties.items():
            result[name] = self._deserialize_primitive(non_ref_prop, hints.get(name))

        for ref_prop in properties.ref_props:
            hint = hints.get(ref_prop.prop_name)
            if hint is not None:
                referenced_property_type = (lambda: "TODO: implement this")()
                result[ref_prop.prop_name] = [
                    _Object(
                        properties=self.__parse_result(
                            prop, cast(Type[Model], referenced_property_type)
                        ),
                        metadata=self._extract_metadata_for_object(prop.metadata)._to_return(),
                    )
                    for prop in ref_prop.properties
                ]
            else:
                raise ValueError(
                    f"Property {ref_prop.prop_name} is not defined with a Reference[Model] type hint in the model {self.model}"
                )

        return type_(**result)

    def __result_to_object(self, res: SearchResult) -> _Object[Model]:
        properties = self.__parse_result(res.properties, self.model)
        metadata = self._extract_metadata_for_object(res.metadata)._to_return()
        return _Object[Model](properties=properties, metadata=metadata)

    def get(
        self,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        after: Optional[UUID] = None,
        filters: Optional[_Filters] = None,
        return_metadata: Optional[MetadataQuery] = None,
        return_properties: Optional[PROPERTIES] = None,
    ) -> List[_Object[Model]]:
        return [
            self.__result_to_object(obj)
            for obj in self._query()
            .get(
                limit=limit,
                offset=offset,
                after=after,
                filters=filters,
                return_metadata=return_metadata,
                return_properties=return_properties,
            )
            .results
        ]

    def hybrid(
        self,
        query: str,
        alpha: Optional[float] = None,
        vector: Optional[List[float]] = None,
        query_properties: Optional[List[str]] = None,
        fusion_type: Optional[HybridFusion] = None,
        limit: Optional[int] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        return_metadata: Optional[MetadataQuery] = None,
        return_properties: Optional[PROPERTIES] = None,
    ) -> List[_Object[Model]]:
        return [
            self.__result_to_object(obj)
            for obj in self._query()
            .hybrid(
                query=query,
                alpha=alpha,
                vector=vector,
                properties=query_properties,
                fusion_type=fusion_type,
                limit=limit,
                autocut=auto_limit,
                filters=filters,
                return_metadata=return_metadata,
                return_properties=return_properties,
            )
            .results
        ]

    def bm25(
        self,
        query: str,
        query_properties: Optional[List[str]] = None,
        limit: Optional[int] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        return_metadata: Optional[MetadataQuery] = None,
        return_properties: Optional[PROPERTIES] = None,
    ) -> List[_Object[Model]]:
        return [
            self.__result_to_object(obj)
            for obj in self._query()
            .bm25(
                query=query,
                properties=query_properties,
                limit=limit,
                autocut=auto_limit,
                filters=filters,
                return_metadata=return_metadata,
                return_properties=return_properties,
            )
            .results
        ]

    def near_vector(
        self,
        near_vector: List[float],
        certainty: Optional[float] = None,
        distance: Optional[float] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        return_metadata: Optional[MetadataQuery] = None,
        return_properties: Optional[PROPERTIES] = None,
    ) -> List[_Object[Model]]:
        return [
            self.__result_to_object(obj)
            for obj in self._query()
            .near_vector(
                near_vector=near_vector,
                certainty=certainty,
                distance=distance,
                autocut=auto_limit,
                filters=filters,
                return_metadata=return_metadata,
                return_properties=return_properties,
            )
            .results
        ]

    def near_object(
        self,
        near_object: UUID,
        certainty: Optional[float] = None,
        distance: Optional[float] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        return_metadata: Optional[MetadataQuery] = None,
        return_properties: Optional[PROPERTIES] = None,
    ) -> List[_Object[Model]]:
        return [
            self.__result_to_object(obj)
            for obj in self._query()
            .near_object(
                near_object=near_object,
                certainty=certainty,
                distance=distance,
                autocut=auto_limit,
                filters=filters,
                return_metadata=return_metadata,
                return_properties=return_properties,
            )
            .results
        ]

    def near_text(
        self,
        query: Union[List[str], str],
        certainty: Optional[float] = None,
        distance: Optional[float] = None,
        move_to: Optional[Move] = None,
        move_away: Optional[Move] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        return_metadata: Optional[MetadataQuery] = None,
        return_properties: Optional[PROPERTIES] = None,
    ) -> List[_Object[Model]]:
        return [
            self.__result_to_object(obj)
            for obj in self._query()
            .near_text(
                near_text=query,
                certainty=certainty,
                distance=distance,
                move_to=move_to,
                move_away=move_away,
                autocut=auto_limit,
                filters=filters,
                return_metadata=return_metadata,
                return_properties=return_properties,
            )
            .results
        ]

    def near_image(
        self,
        near_image: str,
        certainty: Optional[float] = None,
        distance: Optional[float] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        return_metadata: Optional[MetadataQuery] = None,
        return_properties: Optional[PROPERTIES] = None,
    ) -> List[_Object[Model]]:
        return [
            self.__result_to_object(obj)
            for obj in self._query()
            .near_image(
                image=near_image,
                certainty=certainty,
                distance=distance,
                filters=filters,
                autocut=auto_limit,
                return_metadata=return_metadata,
                return_properties=return_properties,
            )
            .results
        ]

    def near_audio(
        self,
        near_audio: str,
        certainty: Optional[float] = None,
        distance: Optional[float] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        return_metadata: Optional[MetadataQuery] = None,
        return_properties: Optional[PROPERTIES] = None,
    ) -> List[_Object[Model]]:
        return [
            self.__result_to_object(obj)
            for obj in self._query()
            .near_audio(
                audio=near_audio,
                certainty=certainty,
                distance=distance,
                filters=filters,
                autocut=auto_limit,
                return_metadata=return_metadata,
                return_properties=return_properties,
            )
            .results
        ]

    def near_video(
        self,
        near_video: str,
        certainty: Optional[float] = None,
        distance: Optional[float] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        return_metadata: Optional[MetadataQuery] = None,
        return_properties: Optional[PROPERTIES] = None,
    ) -> List[_Object[Model]]:
        return [
            self.__result_to_object(obj)
            for obj in self._query()
            .near_video(
                video=near_video,
                certainty=certainty,
                distance=distance,
                filters=filters,
                autocut=auto_limit,
                return_metadata=return_metadata,
                return_properties=return_properties,
            )
            .results
        ]



import unittest
from typing import List, Callable, Tuple
from unittest.mock import patch

from requests.exceptions import ConnectionError as RequestsConnectionError

from test.util import mock_connection_func, check_error_message, check_startswith_error_message
from weaviate.exceptions import UnexpectedStatusCodeException
from weaviate.gql.aggregate import AggregateBuilder


class TestAggregateBuilder(unittest.TestCase):
    def setUp(self):
        self.aggregate = AggregateBuilder("Object", None)

    def test_with_meta_count(self):
        """
        Test the `with_meta_count` method.
        """

        query = self.aggregate.with_meta_count().build()
        self.assertEqual("{Aggregate{Object{meta{count}}}}", query)

    def test_with_fields(self):
        """
        Test the `with_fields` method.
        """

        query = self.aggregate.with_fields("size { mean }").build()
        self.assertEqual("{Aggregate{Object{size { mean }}}}", query)

    def test_with_where(self):
        """
        Test the `with_where` method.
        """

        query = (
            self.aggregate.with_meta_count()
            .with_where({"operator": "Equal", "valueString": "B", "path": ["name"]})
            .build()
        )
        self.assertEqual(
            '{Aggregate{Object(where: {path: ["name"] operator: Equal valueString: "B"} ){meta{count}}}}',
            query,
        )

    def test_group_by_filter(self):
        """
        Test the `with_group_by_filter` method.
        """

        query = (
            self.aggregate.with_group_by_filter(["name"])
            .with_fields("groupedBy { value }")
            .with_fields("name { count }")
            .build()
        )
        self.assertEqual(
            '{Aggregate{Object(groupBy: ["name"]){groupedBy { value }name { count }}}}', query
        )

    def test_with_limit(self):
        """
        Test the `with_limit` method.
        """

        query = (
            self.aggregate.with_meta_count()
            .with_where({"operator": "Equal", "valueString": "B", "path": ["name"]})
            .with_limit(10)
            .build()
        )
        self.assertEqual(
            '{Aggregate{Object(where: {path: ["name"] operator: Equal valueString: "B"} limit: 10){meta{count}}}}',
            query,
        )

    test_near_media_param_list: List[
        Tuple[str, str, str, Callable[[AggregateBuilder, str, str, bool], AggregateBuilder]]
    ] = [
        (
            "audio",
            "test_audio",
            "nearAudio",
            lambda b, k, v, e: b.with_near_audio({k: v, "certainty": 0.55}, encode=e),
        ),
        (
            "video",
            "test_video",
            "nearVideo",
            lambda b, k, v, e: b.with_near_video({k: v, "certainty": 0.55}, encode=e),
        ),
        (
            "depth",
            "test_depth",
            "nearDepth",
            lambda b, k, v, e: b.with_near_depth({k: v, "certainty": 0.55}, encode=e),
        ),
        (
            "thermal",
            "test_thermal",
            "nearThermal",
            lambda b, k, v, e: b.with_near_thermal({k: v, "certainty": 0.55}, encode=e),
        ),
        (
            "imu",
            "test_imu",
            "nearIMU",
            lambda b, k, v, e: b.with_near_imu({k: v, "certainty": 0.55}, encode=e),
        ),
    ]

    def test_near_media(self):
        """
        Test the `with_near_<media>` method.
        """
        for key, value, type_, fn in self.test_near_media_param_list:
            with self.subTest(key=key, value=value):
                with patch(
                    "weaviate.gql.aggregate.file_encoder_b64", side_effect=lambda x: "test_call"
                ) as mocked:
                    # valid calls
                    ## encode False
                    query = fn(
                        AggregateBuilder("Person", None).with_fields("name"), key, value, False
                    ).build()
                    self.assertEqual(
                        f'{{Aggregate{{Person({type_}: {{{key}: "{value}" certainty: 0.55}} ){{name}}}}}}',
                        query,
                    )
                    mocked.assert_not_called()

                    ## encode True
                    query = fn(
                        AggregateBuilder("Person", None).with_fields("name"), key, value, True
                    ).build()
                    self.assertEqual(
                        f'{{Aggregate{{Person({type_}: {{{key}: "test_call" certainty: 0.55}} ){{name}}}}}}',
                        query,
                    )
                    mocked.assert_called()

                    # invalid calls
                    near_error_msg = "Cannot use multiple 'near' filters, or a 'near' filter along with a 'ask' filter!"

                    near_text = {
                        "concepts": "computer",
                        "moveTo": {"concepts": ["science"], "force": 0.5},
                    }
                    with self.assertRaises(AttributeError) as error:
                        fn(
                            AggregateBuilder("Person", None)
                            .with_fields("name")
                            .with_near_text(near_text),
                            key,
                            value,
                            True,
                        )
                    check_error_message(self, error, near_error_msg)

    def test_do(self):
        """
        Test the `do` method.
        """

        # test exceptions
        requests_error_message = "Query was not successful."

        # requests.exceptions.ConnectionError
        mock_obj = mock_connection_func("post", side_effect=RequestsConnectionError("Test"))
        self.aggregate._connection = mock_obj
        with self.assertRaises(RequestsConnectionError) as error:
            self.aggregate.do()
        check_error_message(self, error, requests_error_message)

        # weaviate.UnexpectedStatusCodeException
        mock_obj = mock_connection_func("post", status_code=404)
        self.aggregate._connection = mock_obj
        with self.assertRaises(UnexpectedStatusCodeException) as error:
            self.aggregate.do()
        check_startswith_error_message(self, error, "Query was not successful")

        filter_name = {"path": ["name"], "operator": "Equal", "valueString": "B"}

        self.aggregate.with_group_by_filter(["name"]).with_fields(
            "groupedBy { value }"
        ).with_fields("name { count }").with_where(filter_name)
        expected_gql_clause = '{Aggregate{Object(where: {path: ["name"] operator: Equal valueString: "B"} groupBy: ["name"]){groupedBy { value }name { count }}}}'

        mock_obj = mock_connection_func("post", status_code=200, return_json={"status": "OK!"})
        self.aggregate._connection = mock_obj
        self.assertEqual(self.aggregate.do(), {"status": "OK!"})
        mock_obj.post.assert_called_with(
            path="/graphql", weaviate_object={"query": expected_gql_clause}
        )

    def test_uncapitalized_class_name(self):
        """
        Test the uncapitalized class_name.
        """

        aggregate = AggregateBuilder("Test", None)
        self.assertEqual(aggregate._class_name, "Test")

        aggregate = AggregateBuilder("test", None)
        self.assertEqual(aggregate._class_name, "Test")



import unittest
from sys import platform
from unittest.mock import patch, Mock

from requests.exceptions import ConnectionError as RequestsConnectionError

from test.util import mock_connection_func, check_error_message
from weaviate import Client, ConnectionConfig, ConnectionParams, ProtocolParams
from weaviate.embedded import EmbeddedOptions, EmbeddedDB
from weaviate.exceptions import UnexpectedStatusCodeException


@patch("weaviate.client.Connection", Mock)
class TestClient(unittest.TestCase):
    @patch("weaviate.client.Client.get_meta", return_value={"version": "1.13.2"})
    def test___init__(self, mock_get_meta_method):
        """
        Test the `__init__` method.
        """

        type_error_message = "Either url or embedded options must be present."
        # test invalid calls
        with self.assertRaises(TypeError) as error:
            Client(None)
        check_error_message(self, error, type_error_message)
        with self.assertRaises(TypeError) as error:
            Client(42)
        check_error_message(
            self,
            error,
            "URL is expected to be string but is " + str(int),
        )

        # test valid calls
        with patch(
            "weaviate.client.Connection",
            Mock(side_effect=lambda **kwargs: Mock(timeout_config=kwargs["timeout_config"])),
        ) as mock_obj:
            Client(
                url="http://localhost:8080",
                auth_client_secret=None,
                timeout_config=(1, 2),
                additional_headers=None,
                startup_period=None,
            )
            mock_obj.assert_called_with(
                connection_params=ConnectionParams(
                    http=ProtocolParams(
                        host="localhost",
                        port=8080,
                        secure=False,
                    )
                ),
                auth_client_secret=None,
                timeout_config=(1, 2),
                proxies=None,
                trust_env=False,
                additional_headers=None,
                startup_period=None,
                embedded_db=None,
                connection_config=ConnectionConfig(),
            )

        with patch(
            "weaviate.client.Connection",
            Mock(side_effect=lambda **kwargs: Mock(timeout_config=kwargs["timeout_config"])),
        ) as mock_obj:
            Client(
                url="http://localhost:8080",
                auth_client_secret=None,
                timeout_config=(1, 2),
                additional_headers={"Test": True},
                startup_period=None,
            )
            mock_obj.assert_called_with(
                connection_params=ConnectionParams(
                    http=ProtocolParams(
                        host="localhost",
                        port=8080,
                        secure=False,
                    )
                ),
                auth_client_secret=None,
                timeout_config=(1, 2),
                proxies=None,
                trust_env=False,
                additional_headers={"Test": True},
                startup_period=None,
                embedded_db=None,
                connection_config=ConnectionConfig(),
            )

        with patch(
            "weaviate.client.Connection",
            Mock(side_effect=lambda **kwargs: Mock(timeout_config=kwargs["timeout_config"])),
        ) as mock_obj:
            Client(
                url="http://localhost:8080",
                auth_client_secret=None,
                timeout_config=(5, 20),
                startup_period=None,
            )
            mock_obj.assert_called_with(
                connection_params=ConnectionParams(
                    http=ProtocolParams(
                        host="localhost",
                        port=8080,
                        secure=False,
                    )
                ),
                auth_client_secret=None,
                timeout_config=(5, 20),
                proxies=None,
                trust_env=False,
                additional_headers=None,
                startup_period=None,
                embedded_db=None,
                connection_config=ConnectionConfig(),
            )

        with patch(
            "weaviate.client.Connection",
            Mock(side_effect=lambda **kwargs: Mock(timeout_config=kwargs["timeout_config"])),
        ) as mock_obj:
            Client(
                url="http://localhost:8080",
                auth_client_secret=None,
                timeout_config=(1, 2),
                proxies={"http": "test"},
                trust_env=True,
                additional_headers=None,
                startup_period=None,
            )
            mock_obj.assert_called_with(
                connection_params=ConnectionParams(
                    http=ProtocolParams(
                        host="localhost",
                        port=8080,
                        secure=False,
                    )
                ),
                auth_client_secret=None,
                timeout_config=(1, 2),
                proxies={"http": "test"},
                trust_env=True,
                additional_headers=None,
                startup_period=None,
                embedded_db=None,
                connection_config=ConnectionConfig(),
            )

        if platform == "linux":
            with patch(
                "weaviate.client.Connection",
                Mock(side_effect=lambda **kwargs: Mock(timeout_config=kwargs["timeout_config"])),
            ) as mock_obj:
                with patch("weaviate.embedded.EmbeddedDB.start") as mocked_start:
                    Client(embedded_options=EmbeddedOptions())
                    args, kwargs = mock_obj.call_args_list[0]
                    self.assertEqual(
                        kwargs["connection_params"],
                        ConnectionParams.from_url("http://localhost:8079", 50060),
                    )
                    self.assertTrue(isinstance(kwargs["embedded_db"], EmbeddedDB))
                    self.assertTrue(kwargs["embedded_db"] is not None)
                    self.assertEqual(kwargs["embedded_db"].options.port, 8079)
                    mocked_start.assert_called_once()

    @patch("weaviate.client.Client.get_meta", return_value={"version": "1.13.2"})
    def test_is_ready(self, mock_get_meta_method):
        """
        Test the `is_ready` method.
        """
        client = Client(
            url="http://localhost:8080",
        )
        # Request to weaviate returns 200
        connection_mock = mock_connection_func("get")
        client._connection = connection_mock
        self.assertTrue(client.is_ready())  # Should be true
        connection_mock.get.assert_called_with(path="/.well-known/ready")

        # Request to weaviate returns 404
        connection_mock = mock_connection_func("get", status_code=404)
        client._connection = connection_mock
        self.assertFalse(client.is_ready())  # Should be false
        connection_mock.get.assert_called_with(path="/.well-known/ready")

        # Test exception in connect
        connection_mock = mock_connection_func("get", side_effect=RequestsConnectionError("Test"))
        client._connection = connection_mock
        self.assertFalse(client.is_ready())
        connection_mock.get.assert_called_with(path="/.well-known/ready")

    @patch("weaviate.client.Client.get_meta", return_value={"version": "1.13.2"})
    def test_is_live(self, mock_get_meta):
        """
        Test the `is_live` method.
        """
        client = Client("http://localhost:8080")
        # Request to weaviate returns 200
        connection_mock = mock_connection_func("get")
        client._connection = connection_mock
        self.assertTrue(client.is_live())  # Should be true
        connection_mock.get.assert_called_with(path="/.well-known/live")

        # Request to weaviate returns 404
        connection_mock = mock_connection_func("get", status_code=404)
        client._connection = connection_mock
        self.assertFalse(client.is_live())  # Should be false
        connection_mock.get.assert_called_with(path="/.well-known/live")

    def test_get_meta(self):
        """
        Test the `get_meta` method.
        """

        # client = Client("http://localhost:8080")
        # # Request to weaviate returns 200
        # connection_mock = mock_connection_func('get', return_json="OK!")
        # client._connection = connection_mock
        # self.assertEqual(client.get_meta(), "OK!")
        # connection_mock.get.assert_called_with(
        #     path="/meta"
        # )

        # # Request to weaviate returns 404
        # connection_mock = mock_connection_func('get', status_code=404)
        # client._connection = connection_mock
        # with self.assertRaises(UnexpectedStatusCodeException) as error:
        #     client.get_meta()
        # error_message = "Meta endpoint! Unexpected status code: 404, with response body: None"
        # check_error_message(self, error, error_message)
        # connection_mock.get.assert_called_with(
        #     path="/meta"
        # )

    @patch("weaviate.client.Client.get_meta", return_value={"version": "1.13.2"})
    def test_get_open_id_configuration(self, mock_get_meta):
        """
        Test the `get_open_id_configuration` method.
        """
        client = Client(
            url="http://localhost:8080",
        )
        # Request to weaviate returns 200
        connection_mock = mock_connection_func("get", return_json={"status": "OK!"})
        client._connection = connection_mock
        self.assertEqual(client.get_open_id_configuration(), {"status": "OK!"})
        connection_mock.get.assert_called_with(path="/.well-known/openid-configuration")

        # Request to weaviate returns 404
        connection_mock = mock_connection_func("get", status_code=404)
        client._connection = connection_mock
        self.assertIsNone(client.get_open_id_configuration())
        connection_mock.get.assert_called_with(path="/.well-known/openid-configuration")

        # Request to weaviate returns 204
        connection_mock = mock_connection_func("get", status_code=204)
        client._connection = connection_mock
        with self.assertRaises(UnexpectedStatusCodeException) as error:
            client.get_open_id_configuration()
        error_message = "Meta endpoint! Unexpected status code: 204, with response body: None."
        check_error_message(self, error, error_message)
        connection_mock.get.assert_called_with(path="/.well-known/openid-configuration")

    @patch("weaviate.client.Client.get_meta", return_value={"version": "1.13.2"})
    def test_timeout_config(self, mock_get_meta):
        """
        Test the `set_timeout_config` method.
        """
        client = Client(
            url="http://some_url.com:80", auth_client_secret=None, timeout_config=(1, 2)
        )
        self.assertEqual(client.timeout_config, (1, 2))
        client.timeout_config = (4, 20)  # ;)
        self.assertEqual(client.timeout_config, (4, 20))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: v1/base.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\rv1/base.proto\x12\x0bweaviate.v1\x1a\x1cgoogle/protobuf/struct.proto":\n\x15NumberArrayProperties\x12\x0e\n\x06values\x18\x01 \x03(\x01\x12\x11\n\tprop_name\x18\x02 \x01(\t"7\n\x12IntArrayProperties\x12\x0e\n\x06values\x18\x01 \x03(\x03\x12\x11\n\tprop_name\x18\x02 \x01(\t"8\n\x13TextArrayProperties\x12\x0e\n\x06values\x18\x01 \x03(\t\x12\x11\n\tprop_name\x18\x02 \x01(\t";\n\x16\x42ooleanArrayProperties\x12\x0e\n\x06values\x18\x01 \x03(\x08\x12\x11\n\tprop_name\x18\x02 \x01(\t"\xd7\x03\n\x15ObjectPropertiesValue\x12\x33\n\x12non_ref_properties\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x43\n\x17number_array_properties\x18\x02 \x03(\x0b\x32".weaviate.v1.NumberArrayProperties\x12=\n\x14int_array_properties\x18\x03 \x03(\x0b\x32\x1f.weaviate.v1.IntArrayProperties\x12?\n\x15text_array_properties\x18\x04 \x03(\x0b\x32 .weaviate.v1.TextArrayProperties\x12\x45\n\x18\x62oolean_array_properties\x18\x05 \x03(\x0b\x32#.weaviate.v1.BooleanArrayProperties\x12\x38\n\x11object_properties\x18\x06 \x03(\x0b\x32\x1d.weaviate.v1.ObjectProperties\x12\x43\n\x17object_array_properties\x18\x07 \x03(\x0b\x32".weaviate.v1.ObjectArrayProperties"^\n\x15ObjectArrayProperties\x12\x32\n\x06values\x18\x01 \x03(\x0b\x32".weaviate.v1.ObjectPropertiesValue\x12\x11\n\tprop_name\x18\x02 \x01(\t"X\n\x10ObjectProperties\x12\x31\n\x05value\x18\x01 \x01(\x0b\x32".weaviate.v1.ObjectPropertiesValue\x12\x11\n\tprop_name\x18\x02 \x01(\t*\x89\x01\n\x10\x43onsistencyLevel\x12!\n\x1d\x43ONSISTENCY_LEVEL_UNSPECIFIED\x10\x00\x12\x19\n\x15\x43ONSISTENCY_LEVEL_ONE\x10\x01\x12\x1c\n\x18\x43ONSISTENCY_LEVEL_QUORUM\x10\x02\x12\x19\n\x15\x43ONSISTENCY_LEVEL_ALL\x10\x03\x42n\n#io.weaviate.client.grpc.protocol.v1B\x11WeaviateProtoBaseZ4github.com/weaviate/weaviate/grpc/generated;protocolb\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "v1.base_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b"\n#io.weaviate.client.grpc.protocol.v1B\021WeaviateProtoBaseZ4github.com/weaviate/weaviate/grpc/generated;protocol"
    _globals["_CONSISTENCYLEVEL"]._serialized_start = 957
    _globals["_CONSISTENCYLEVEL"]._serialized_end = 1094
    _globals["_NUMBERARRAYPROPERTIES"]._serialized_start = 60
    _globals["_NUMBERARRAYPROPERTIES"]._serialized_end = 118
    _globals["_INTARRAYPROPERTIES"]._serialized_start = 120
    _globals["_INTARRAYPROPERTIES"]._serialized_end = 175
    _globals["_TEXTARRAYPROPERTIES"]._serialized_start = 177
    _globals["_TEXTARRAYPROPERTIES"]._serialized_end = 233
    _globals["_BOOLEANARRAYPROPERTIES"]._serialized_start = 235
    _globals["_BOOLEANARRAYPROPERTIES"]._serialized_end = 294
    _globals["_OBJECTPROPERTIESVALUE"]._serialized_start = 297
    _globals["_OBJECTPROPERTIESVALUE"]._serialized_end = 768
    _globals["_OBJECTARRAYPROPERTIES"]._serialized_start = 770
    _globals["_OBJECTARRAYPROPERTIES"]._serialized_end = 864
    _globals["_OBJECTPROPERTIES"]._serialized_start = 866
    _globals["_OBJECTPROPERTIES"]._serialized_end = 954
# @@protoc_insertion_point(module_scope)
